<?php
/**
 * General mail access wich is accessed thru a file
 * instead of a string. This is done to handle files
 * which would exeed the memory_limit of php. Off Course
 * this limit can be disabled, but this should be a more
 * smooth solution.
 *
 * Copyright 2005      Auke Bruinsma <a.bruinsma@tools4ever.com>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 *
 * @author  Auke H. Bruinsma <a.bruinsma@tools4ever.com>
 * @package Whups
 */

/*
if (!function_exists ('stripos')) {
    function stripos($haystack,$needle,$offset = 0)
    {
       return(strpos(strtolower($haystack),strtolower($needle),$offset));
    }
}

/*
if (!function_exists('mime_content_type')) {
   function mime_content_type($f) {
       $f = escapeshellarg($f);
       return trim( `file -bi $f` );
   }
}
*/

/*
if (!function_exists('v_dump')) {
    function v_dump($var, $exit = false)
    {
        print '<pre>';
        var_dump($var);
        print '</pre>';

        if (true == $exit) {
            exit;
        }
    }
}
*/

        function implode_assoc_r ($array = null, $inner_glue = "=", $outer_glue = "\r\n",  $keepOuterKey = false)
        {
           $output = array();

           foreach( $array as $key => $item )
               if ( is_array ($item) )
               {
                   if ( $keepOuterKey )
                       $output[] = $key;

                   // This is value is an array, go and do it again!
                   $output[] = implode_assoc_r ($item, $inner_glue, $outer_glue, $keepOuterKey);
               }
               else
                   $output[] = $key . $inner_glue . $item;

           return implode($outer_glue, $output);
        }


class File_Mail {
    var $_body;
    var $_file;
    var $_header;
    var $_disposition;
    var $_startpos;
    var $_parts = array ();
    var $_type = NULL;
    var $_include_bodies = true;
    var $_decode_bodies = true;
    var $_attachthebody;
    var $_processed = false;
    var $_contentTypeParameters = NULL;
    /**
     * Parse a MIME message
     *
     * @param string $file              This is the full path of the file
     *                                  that contains the MIME message.
     */
    function File_Mail ($file)
    {
        register_shutdown_function (array( &$this, "File_Mail_Destory" ));
        $this->_file = $file;
        $this->_header = NULL;
        $this->_startpos = 0;
        $this->GetHeaders ();
    }
    
    function File_Mail_Destory ()
    {
        if (isset ($this->file)) if (file_exists ($this->file)) unlink ($this->file); 
        if (isset ($this->_body)) unset ($this->_body);
    }
 
    function GetHeaders ()
    {
        if (strlen ($this->_header) > 0) return $this->_header;
        return $this->_RefreshHeader ();
    }

    function _getLine (&$handle)
    {
        $line = "";
        $char = "";
        $newline = false;
        $stop = 0;
        $counter = 0;
        while (!feof($handle)) {
            $char = fgetc ($handle);
            if ($char === false) break;
            if ($counter > 2048) break;
            $line .= $char;
            $counter++;
            if ($newline && ($char != "\r" && $char != "\n")) {
                break;
            }
//            $line .= $char;
            if ($char == "\n" || $char == "\r") $newline = true;
            else ($newline = false);
        }
        return  $line;
    }

    /**
     * Reads the header from the assigned file.
     * 
     * @return string The header of the file.
     * @access private
     */
    function _RefreshHeader ()
    {
        $mailhandle =  fopen($this->_file, "rb");
        if (!$mailhandle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed."), 'horde.error');
        }
        $this->_header = "";
        while (!feof($mailhandle)) {
               $newblock = $this->_getLine ($mailhandle);
               if (!$this->_AddHeaderBlock ($newblock)) break;
        }
        $this->_startpos = ftell ($mailhandle);
	fclose($mailhandle);
        return $this->_header;
    }

    function _boundarySplit ($id)
    {
        $mailhandle = fopen($this->_file, "rb");
        if (!$mailhandle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed.")
, 'horde.error');
        }


        fseek ($mailhandle, $this->_startpos);
        $positions = array ();
        while (!feof($mailhandle)) {
               $prefpos = ftell ($mailhandle) - 1;
               $newblock = $this->_getLine ($mailhandle);
               if (strpos ($newblock, $id) !== false)
               {
                    $positions [] = $prefpos;
                    $positions [] = ftell ($mailhandle) - 1;
               }
 	}
	array_shift($positions);
        array_pop($positions);
        for ($i = 0; $i < (count ($positions) - 1); $i+=2) {
            if ($positions [$i] == $positions [$i + 1]) {
               continue;
            }
            $filename = tempnam ("/tmp", "whupsmessagepart");
            $handle = fopen($filename, "wb");
            fseek ($mailhandle, $positions [$i]);
            if (!$handle) {
                return PEAR::raiseError(_("The given file " . $filename . " cannot be opened. The e-mail cannot be processed."), 'horde.error');
            }
            $counter = $positions [$i];
            while (!feof ($mailhandle) && $counter < $positions [$i + 1]) {
                 fwrite ($handle, fgetc ($mailhandle)); $counter++;
            }
            fclose ($handle);
            $part = &new File_Mail ($filename);
            $this->_parts [] = &$part; // new File_Mail ($handle);
              
        } 
        fclose ($mailhandle);
    }
 
    function _AddHeaderBlock ($newblock)
    {
        if (preg_match ("/^(.*?)\r?\n\r?\n/s", $newblock, $matches)) {
            // this block contains the header and, so we add
            // the header part and return false;
            $this->_header .= $matches[1];
            return false;
        } 
        else {
            // this block does not contain the end of the headers
            // so add it and return true. 
            $this->_header .= $newblock;
            return true;
        }
    }

    function _AddAttachementBlock ($handle, $newblock)
    {
        if (preg_match ("/^(.*?)\r?\n\r?\n/s", $newblock, $matches)) {
            // this block contains the header and, so we add
            // the header part and return false;
            fwrite ($handle, $matches[1]);
            return false;
        }
        else {
            // this block does not contain the end of the headers
            // so add it and return true.
            fwrite ($handle, $newblock);
            return true;
        }
    }
 
    function getContents () 
    {   
        if ($this->_body === NULL) $this->getType ();
                     
        return $this->_body;
    }

    function ParseType ()
    {
         if ($this->_processed) 
	 {
            // Should not get here, but in case it does...
            // If _processed is true, the type should already be know and this
            // function may not be called again, because the associated files are already
            // decoded. 
            // They cannot be decoded again. (they are decoded in _ParseType which is called
            // from this function.
            return $this->_type;
         }
         $return = new stdClass;
         $headers = MIME_Structure::parseMIMEHeaders($this->GetHeaders ());      	
         $this->_attachthebody = false;

 	 foreach ($headers as $key => $value) {
             $headers [strtolower ($key)] = $value;
             switch (strtolower ($key)) {
                 case 'content-type':
                    $content_type = $this->_parseHeaderValue ($value);
                    if (preg_match('/([0-9a-z+.-]+)\/([0-9a-z+.-]+)/i', $content_type['value'], $regs)) {
                        $return->ctype_primary   = $regs[1];
                        $return->ctype_secondary = $regs[2];
                    }

                    if (isset($content_type['other'])) {
                        while (list($p_name, $p_value) = each($content_type['other'])) {
                            $return->ctype_parameters[$p_name] = $p_value;
                            $this->_contentTypeParameters [$p_name] =  $p_value;
                        }
                    }
                    break;

                case 'content-disposition';
                    $content_disposition = $this->_parseHeaderValue($value);
                    $this->_disposition   = $content_disposition['value'];
                    if (isset($content_disposition['other'])) {
                        while (list($p_name, $p_value) = each($content_disposition['other'])) {
                            $return->d_parameters[$p_name] = $p_value;
                            $this->_contentTypeParameters [$p_name] =  $p_value;

                        }
                    }
                    break;

                case 'content-transfer-encoding':
                    $content_transfer_encoding = $this->_parseHeaderValue($value);
                    break;
             } 
         }

        if (isset($content_type)) {
            switch (strtolower($content_type['value'])) {
                case 'text/plain':
                    $encoding = isset($content_transfer_encoding) ? $content_transfer_encoding['value'] : '7bit';
                    $this->_include_bodies ? $return->body = ($this->_decode_bodies ? $this->_decodeBody($encoding) : $body) : null;
                    break;

                case 'text/html':
                    $encoding = isset($content_transfer_encoding) ? $content_transfer_encoding['value'] : '7bit';
                    $this->_include_bodies ? $return->body = ($this->_decode_bodies ? $this->_decodeBody($encoding) : $body) : null;
                    break;

                case 'multipart/parallel':
                case 'multipart/report': // RFC1892
                case 'multipart/signed': // PGP
                case 'multipart/digest':
                case 'multipart/alternative':
                case 'multipart/related':
                case 'multipart/mixed':
                    if(!isset($content_type['other']['boundary'])){
                        $this->_error = 'No boundary found for ' . $content_type['value'] . ' part';
                        return false;
                    }

                    $default_ctype = (strtolower($content_type['value']) === 'multipart/digest') ? 'message/rfc822' : 'text/plain';
                    		    
                    //$parts = 
                    $this->_boundarySplit ($content_type['other']['boundary']);
                    break;

                case 'message/rfc822':
                    $obj = &new Mail_mimeDecode($body);
                    $this->_parts[] = $obj->decode(array('include_bodies' => $this->_include_bodies));
                    unset($obj);
                    break;

                default:
                    if(!isset($content_transfer_encoding['value']))
                        $content_transfer_encoding['value'] = '7bit';
                    $this->_attachthebody = true;
                    $this->_include_bodies ? $this->_bodyfile = ($this->_decode_bodies ? $this->_decodeBody ($content_transfer_encoding['value']) : "") : null;
                    break;
            }

        } 
        else {
            $default_ctype = 'text/plain';
            $ctype = explode('/', $default_ctype);
            $return->ctype_primary   = $ctype[0];
            $return->ctype_secondary = $ctype[1];
        }
        if (isset ($return->ctype_primary) && isset ($return->ctype_secondary))
        {
            $this->_type = implode ('/', array ($return->ctype_primary, $return->ctype_secondary));
        }
        else 
        {
            $this->_type = $return; //implode ('/', array ($return->ctype_primary, $return->ctype_secondary));
        }
        $this->_processed = true;
        return $this->_type;

    }

    /**
     * Return the type of this message part.
     *
     * @return string The type of this message part.
     * @access public
     */
    function getType () {
        if ($this->_type === NULL) return $this->ParseType ();
        return $this->_type;
    }

    /**
     * This function will decode the file that is assigned to this mail part.
     *
     * @access private
     */
    function _base64decode ()
    {
        $filename = tempnam ("/tmp", "whupsdecodedpart");
        $handle =  fopen($filename, "wb");
        if (!$handle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed.")
, 'horde.error');
        }

        $mailhandle =  fopen($this->_file, "rb");
        if (!$mailhandle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed.")
, 'horde.error');
        }
        fseek ($mailhandle, $this->_startpos - 1);

        while (!feof($mailhandle)) {
               $newblock = rtrim (fgets ($mailhandle, 5), '=');
               fwrite ($handle, base64_decode ($newblock));
        }
        fclose($mailhandle);
        fclose ($handle);
        unlink ($this->_file);
        $this->_file = $filename;
        return;
    }

    /**
     * This function will decode the file that is assigned to this mail part.
     *
     * @return string The first 5 kb of the decoded body
     * @access private
     */
    function _quotedPrintableDecode()
    {
        // Remove soft line breaks
        
        $filename = tempnam ("/tmp", "whupsdecodedpart");
        $handle =  fopen($filename, "wb");
        if (!$handle) {
            return PEAR::raiseError(_("The given file " . $filename . " cannot be opened. The e-mail cannot be processed.")
, 'horde.error');
        }

        $mailhandle =  fopen($this->_file, "rb");
        if (!$mailhandle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed.")
, 'horde.error');
        }
        fseek ($mailhandle, $this->_startpos - 1);
        $bodyblock = "";

        while (!feof($mailhandle)) {
               $newblock = fgets ($mailhandle, 5);
               $newblock = preg_replace("/=\r?\n/", '', $newblock);
               $newblock = preg_replace('/=([a-f0-9]{2})/ie', "chr(hexdec('\\1'))", $newblock);
               if (strlen ($bodyblock < 5120)) $bodyblock .= $newblock;
               fwrite ($handle, base64_decode ($newblock));
        }
        fclose ($mailhandle);
        fclose ($handle);
        unlink ($this->_file);
        $this->_file = $filename;
        return;

        // Replace encoded characters
        $input = preg_replace('/=([a-f0-9]{2})/ie', "chr(hexdec('\\1'))", $input);

        return $input;
    }

    /**
     * This function will return only the given first part of the assigned 
     * file. At default this is 5kb. 
     *
     * @param  int    The size in bytes of the part to return.
     * @return string The first given bytes of the body.
     * @access private
     */
    function _getBodyPart ($partsize=5120)
    {
        $filename = tempnam ("/tmp", "whupsdecodedbody");
        $handle =  fopen($filename, "wb");
        if (!$handle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed.")
, 'horde.error');
        }

        if (!file_exists ($this->_file)) {
           echo (implode_assoc_r (debug_backtrace())); exit (0); }
   
        $mailhandle =  fopen($this->_file, "rb");
        if (!$mailhandle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed.")
, 'horde.error');
        }
        fseek ($mailhandle, $this->_startpos - 1);

        $bodyblock = "";
        while (!feof($mailhandle)) {
               $newblock = fgets ($mailhandle, 512);
               if (strlen ($bodyblock < $partsize)) $bodyblock .= $newblock;
               fwrite ($handle, $newblock);
        }
        fclose ($mailhandle);
        fclose ($handle);
        unlink ($this->_file);
        $this->_file = $filename;
        return $bodyblock;
    }

    /**
     * Given an encoding type, this function will decode the message part and 
     * in case its not base64 return the first 5kb of that part.
     *
     * @param  string Encoding type to use.
     * @return string Decoded body
     * @access private
     */
    function _decodeBody($encoding = '7bit')
    {
        switch ($encoding) {
            case '7bit':
                $this->_body = $this->_getBodyPart ();
                return $this->_body;
                break;

            case 'quoted-printable':
                $this->_body = $this->_quotedPrintableDecode($input);
                return $this->_body;
                break;

            case 'base64':
                return $this->_base64decode();
                break;

            default:
                $this->_body = $this->_getBodyPart ();
                return $this->_body;
        }
    }

    /**
     * Function to parse a header value,
     * extract first part, and any secondary
     * parts (after ;) This function is not as
     * robust as it could be. Eg. header comments
     * in the wrong place will probably break it.
     *
     * @param string Header value to parse
     * @return array Contains parsed result
     * @access private
     */
    function _parseHeaderValue($input)
    {

        if (($pos = strpos($input, ';')) !== false) {

            $return['value'] = trim(substr($input, 0, $pos));
            $input = trim(substr($input, $pos+1));

            if (strlen($input) > 0) {

                // This splits on a semi-colon, if there's no preceeding backslash
                // Can't handle if it's in double quotes however. (Of course anyone
                // sending that needs a good slap).
                $parameters = preg_split('/\s*(?<!\\\\);\s*/i', $input);

                for ($i = 0; $i < count($parameters); $i++) {
                    $param_name  = substr($parameters[$i], 0, $pos = strpos($parameters[$i], '='));
                    $param_value = substr($parameters[$i], $pos + 1);
                    if ($param_value[0] == '"') {
                        $param_value = substr($param_value, 1, -1);
                    }
                    $return['other'][$param_name] = $param_value;
                    $return['other'][strtolower($param_name)] = $param_value;
                }
            }
        } else {
            $return['value'] = trim($input);
        }

        return $return;
    }

    /**
     * Write the body of this message part to a file, the headers will be 
     * saved in the _header variable.
     *
     * @param filename string The filename to write the body of this part to.
     * @access private
     */
    function _writToFile ($filename)
    {
        $filename = tempnam ("/tmp", "wupsattach");
        $handle = fopen($filename, "wb");
        $mailhandle =  fopen($this->_file, "rb");
        if (!$mailhandle) {
            return PEAR::raiseError(_("The given file " . $this->_file . " cannot be opened. The e-mail cannot be processed."), 'horde.error');
        }
        fseek ($mailhandle, $this->_startpos);
        $newblock = $this->_getLine ($mailhandle);

        while (!feof($mailhandle)) {
               $newblock = $this->_getLine ($mailhandle);
               if (!$this->_AddAttachementBlock ($handle, $newblock)) break;
        }
        fclose ($mailhandle);
        fclose ($handle);
        return;
    }

    function splitContents()
    {
        if (!($boundary = $this->getContentTypeParameter('boundary'))) {
            return false;
        }

        if (!strlen($this->_contents)) {
            return false;
        }

        $eol = $this->getEOL();
        $retvalue = false;

        foreach (explode($eol, $this->_contents) as $line) {
            $pos = strpos($line, '--' . $boundary);
            if (($pos === false) && isset($part_ptr)) {
                $data[] = $line;
            } elseif ($pos === 0) {
                $retvalue = true;
                if (isset($part_ptr)) {
                    $this->_parts[$part_ptr]->setContents(implode($eol, $data));
                    $this->_parts[$part_ptr]->splitContents();
                    next($this->_parts);
                } else {
                    reset($this->_parts);
                }
                if (isset($data)) {
                    unset($data);
                }
                $data = array();
                $part_ptr = key($this->_parts);
            }
        }

        return $retvalue;
    }

    function &getParts ()
    {
       return $this->_parts; 
    }

    /**
     * Parse a MIME message and create a new ticket.
     *
     * @param string $text              This is the full text of the MIME
     *                                  message.
     * @param array $info               An array of information for the new
     *                                  ticket.  This should include:
     *                                      'queue' => queue id
     *                                      'type' => type id
     *                                      'state' => state id
     *                                      'priority' => priority id
     * @param optional string $user     This will be the ticket's requestor.
     *                                  If null, we will try to deduce from
     *                                  the message's "From" field--we do
     *                                  NOT default to Auth::getAuth().
     *
     * @returns mixed ? or PEAR_Error:: instance.
     */ /*
    function processMail($text, $info, $user = null)
    {
        global $conf, $whups;

        require_once 'Horde/MIME/Structure.php';
        $message = &MIME_Structure::parseTextMIMEMessage($text);

        if (preg_match("/^(.*?)\r?\n\r?\n/s", $text, $matches)) {
            $hdrText = $matches[1];
        } else {
            $hdrText = $text;
        }
        $headers = MIME_Structure::parseMIMEHeaders($hdrText);
        $isspam = Whups_Mail::_findSimpleHeaderValue($headers, 'X-Spam-Flag');
        if ($isspam == "YES") 
        {
            unset ($info['queue']);
            $queues = $whups->getQueues();
            foreach ($queues as $queueId => $queueName) {
                if (strcasecmp($queueName, "Spam") == 0) {
                        $info['queue'] = $queueId;
                        break;
                }
            }
            if (!isset($info['queue'])) {
                exit ("The spam queue does not exists... The message cannot be delivered.");
            }
           echo ("Spam e-mail");
        }
        // Use the message subject as the ticket summary.
        $info['summary'] = Whups_Mail::_findSimpleHeaderValue($headers,
                                                              'Subject');

        if (is_null($user)) {
            $user = Whups_Mail::_findAuthUser($headers);
        }

        if (!is_null($user)) {
            if ($user != Auth::getAuth()) {
                $params = Horde::getDriverConfig('auth', $conf['auth']['driver']);
                $auth = &Auth::singleton($conf['auth']['driver'], $params);
                $auth->setAuth($user, array());
            }
            // FIXME: Use Identity to set email address.
        } else {
            $from = Whups_Mail::_findSimpleHeaderValue($headers, 'From');
            if (!empty($from)) {
                $info['user_email'] = $from;
            }
        }
	$ticket = Whups_Mail::_getReplyTicket ($info['summary']);
	if ($ticket) {
            $info['comment'] = Whups_Mail::_buildComment($message, $headers, $ticket);
	    $ticket->change ('comment', substr ($info ['comment'],0, stripos ($info ['comment'], 'Received message:', 18)));
            $ticket->commit ("Air2");
	    return $ticket;
        } else {
             $ticket = "";
             $info['comment'] = Whups_Mail::_buildComment($message, $headers, $ticket);
            $ticket = &Whups_Ticket::newTicket($info, $user);
            $info['comment'] = Whups_Mail::_buildComment($message, $headers, $ticket);
            return $ticket; 
        }
         
    }


    function _findSimpleHeaderValue($headers, $name)
    {
        foreach ($headers as $hdrName => $hdrVals) {
            if (strcasecmp($hdrName, $name) == 0) {
                if (is_array($hdrVals)) {
                    return implode(', ', $hdrVals);
                } else {
                    return $hdrVals;
                }
            }
        }

        return null;
    }

    function &_getReplyTicket($subject)
    {
	//FIXME: Should not only check for Re: but also for Aw: (germany) etc.
        //DONE: ?? Well just don't check for any RE Aw etc. Just find that ticket number.
	//if(stripos($subject, 'RE:') !== false) {
            $ticketnum = stripos($subject, 'Tickets #') + 9;
            if($ticketnum === false) return NULL;
            $ticketnumend = stripos($subject, ']', $ticketnum);
            if($ticketnumend === false) return NULL;
            $ticketnum = substr($subject, $ticketnum, $ticketnumend - $ticketnum);
            echo "ticketnummer:\n";
            $ticket = &Whups_Ticket::makeTicket($ticketnum);
            if (is_a($ticket, 'PEAR_Error')) {
            if ($ticket->code === 0) {
                // No permissions to this ticket.
                echo ($ticket->getMessage());
                return NULL;
            } else {
                echo ($ticket->getMessage());
                return NULL;
            }
	    }
            return $ticket;
    }
    
    /**
     * Search the 'From' header for an email address contained in one
     * of our users' identities.
     *
     * @access private
     */
/*
    function _findAuthUser($headers)
    {
        global $conf;

        require_once 'Horde/Identity.php';
        require_once 'Horde/MIME.php';

        $from = Whups_Mail::_findSimpleHeaderValue($headers, 'From');
        $from = MIME::bareAddress($from);

        $params = Horde::getDriverConfig('auth', $conf['auth']['driver']);
        $auth = &Auth::singleton($conf['auth']['driver'], $params);
        $users = $auth->listUsers();
        foreach ($users as $user) {
            $identity = &Identity::singleton('none', $user);
            $addrs = $identity->getAll('from_addr');
            foreach ($addrs as $addr) {
                if (strcasecmp($from, $addr) == 0) {
                    return $user;
                }
            }
        }

        return null;
    }
*/
}
